<?php
/*

This is PHPAuthentication v1.0, a PHP implementation of password login.

Features:

- no cleartext password through the network, either in POST or in GET parms
- stateless: no database or disk file needed to keep track of CHAP negotiation
- no credentials replay is possible, unless using same IP, same browser 
  version and same platform before the cookie expires
- self-contained: no external libraries needed
- response to challenge is passed through POST: no track in server's logs

Caveats:

- the browser must have javascript and cookies enabled
- the password shows in cleartext in the PHP script
- users cannot change the password
- a change of IP address or browser causes session abort
- logging out wipes the cookie off the client, but the session data will be
  accepted by the server to the full configured lifetime.

Usage:

- put the following code at the beginning of the PHP program:
  require "PHPAuthentication.php";
  $auth = new PHPAuthentication('my_secret_password');
  $priv = $auth->check_privileges();

- insert the following code in the HTML <header>:
  <?php $auth->javascript_login_function() ?>

- in the body: 
  * put the following code where you want a link to login/logout interface:
    <?php $auth->login_logout_link($priv) ?>
  * use the boolean $priv previously defined as needed: $priv is true if 
    user is logged in, false otherwise

Copyright  2009 Alberto Longhi <al@regoloarmonico.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details. 

For more information on the GNU General Public License please see 
http://www.gnu.org/licenses/. 

 */

class PHPAuthentication {
  var $password         = '';
  var $chaptimeout      = 30;                  // seconds
  var $session_lifetime = 3600;                // seconds
  var $cookiename       = 'PHPAuthentication_COOKIE';
  var $parname          = 'PHPAuthentication'; // POST/GET parameter to use
  var $funcname         = 'PHPAuthentication'; // javascript login function
  var $failure_reason   = '';

  function PHPAuthentication($pwd,$slt=null,$cto=null,$cnm=null,$pnm=null){
    $this->password = $pwd;
    if($slt) $this->session_lifetime = $slt;
    if($cto) $this->chaptimeout;
    if($cnm) $this->cookiename = $cnm;
    if($pnm) $this->parname = $pnm;
  }
  function login_logout_link($priv,$intxt='login',$outtxt='logout'){
    if(!$priv)
      echo "<a href=\"javascript:".$this->funcname."('login')\">$intxt</a>";
    else
      echo "<a href=\"javascript:".$this->funcname."('logout')\">$outtxt</a>";
  }
  function challenge($t=null){
    $z      = sprintf("%d",$t?$t:time());
    $token  = $_SERVER["HTTP_USER_AGENT"];
    $token .= $_SERVER["REMOTE_ADDR"];
    $token .= $_SERVER["HTTP_VIA"];
    $token .= $_SERVER["HTTP_X_FORWARDED_FOR"];
    return md5("$token$z");
  }
  function response($t=null){
    return md5($this->challenge($t).$this->password);
  }
  function verify($response){
    $t = time();
    for($i=0;$i<$this->chaptimeout;$i++)
      if($this->response($t-$i)===$response)
	return $t-$i;
    return false;
  }
  function set_cookie($t,$response){
    setcookie($this->cookiename,"$t+$response",time()+$this->session_lifetime);
  }
  function reset_cookie(){
    setcookie($this->cookiename,$_COOKIE[$this->cookiename],time()-9999);
  }
  function verify_cookie(){
    $a=explode("+", $_COOKIE[$this->cookiename]);
    $c=array('t'=>intval($a[0]),'response'=>$a[1]);
    if($this->response($c['t'])===$c['response'])
      return true;
    else
      return false;
  }
  function check_privileges(){
    if(isset($_POST[$this->parname]))
      $response = $_POST[$this->parname];
    elseif(isset($_REQUEST[$this->parname]))
      $response = $_REQUEST[$this->parname];
    else
      $response = false;	
    if($response){
      // we are in the little login/logout window
      if($response=='login'){
	// logging in
	$privileged = false;
      } elseif($response=='logout'){
	// logged out
	$this->reset_cookie();
	$privileged = false;
	$this->failure_reason='logged out';
      } else {
	$t=$this->verify($response);
	if($t){
	  // login successful
	  $this->set_cookie($t,$response);
	  $privileged = true;
	} else {
	  // password incorrect or challenge expired
	  $this->reset_cookie();
	  $privileged = false;
	  $this->failure_reason='password incorrect or challenge expired';
	}
      }
      print $this->loginpage($privileged);
      exit;
      // no more content to login/logout window
    } elseif(isset($_COOKIE[$this->cookiename])){
      if($this->verify_cookie()){
	// open session
	$privileged = true;
      } else {
	// session expired or stale cookie
	$this->reset_cookie();
	$privileged = false;
	$this->failure_reason='session expired or stale cookie';
      }
    } else {
      // no credentials
      $privileged = false;
      $this->failure_reason='no credentials provided';
    }
    return $privileged;
  }
  function loginpage($priv){
?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" 
   "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" lang="en" xml:lang="en">
<head>
<title>PHPAuthentication login</title>
<script type="text/javascript"><?php $this->javascript_chap_functions() ?></script>
</head>
<body>
<?php if(!$priv){ ?>
<?php if($this->failure_reason) echo '<p>'.$this->failure_reason.'</p>' ?>
<form action="" method="post" onsubmit="login()">
<p>To log in, enter the password within <?php echo $this->chaptimeout ?> 
seconds from now, or it will not be accepted.</p>
<p>
<input type="password" name="password" />
<input type="hidden" name="<?php echo $this->parname ?>" value="" />
<input type="submit" value="login" />
</p>
</form>
<?php } else { ?>
<p>You are logged in. You may need to refresh the main window now.<br/>
<a href="?<?php echo $this->parname ?>=logout">logout</a></p>
<?php } ?>
<p><a href="javascript:window.close()">close</a></p>
</body>
</html>
<?php
  }
  function javascript_login_function(){
    echo "
<script type=\"text/javascript\">
function ". $this->funcname ."(action){
  url=''+self.location.pathname+'?". $this->parname ."='+action;
  features = 'toolbar=0,'+
             'location=0,'+
             'directories=0,'+
             'personalbar=0,'+
             'status=0,'+
             'menubar=0,'+
             'dependent=1,'+
             'alwaysRaised=1,'+
             'width=300,'+
             'height=200';
  loginwin = window.open(url,'loginwin',features,false);
}
</script>
";
  }
  function javascript_chap_functions(){
    echo "
//  MD5 algorithm from http://pajhome.org.uk/crypt/md5/md5src.html
var chrsz   = 8;
var hexcase = 0;
function md5(s){
  return binl2hex(core_md5(str2binl(s), s.length * chrsz));
}
function core_md5(x, len){
  x[len >> 5] |= 0x80 << ((len) % 32);
  x[(((len + 64) >>> 9) << 4) + 14] = len;
  var a =  1732584193;
  var b = -271733879;
  var c = -1732584194;
  var d =  271733878;
  for(var i = 0; i < x.length; i += 16){
    var olda = a;
    var oldb = b;
    var oldc = c;
    var oldd = d;
    a = md5_ff(a, b, c, d, x[i+ 0], 7 , -680876936);
    d = md5_ff(d, a, b, c, x[i+ 1], 12, -389564586);
    c = md5_ff(c, d, a, b, x[i+ 2], 17,  606105819);
    b = md5_ff(b, c, d, a, x[i+ 3], 22, -1044525330);
    a = md5_ff(a, b, c, d, x[i+ 4], 7 , -176418897);
    d = md5_ff(d, a, b, c, x[i+ 5], 12,  1200080426);
    c = md5_ff(c, d, a, b, x[i+ 6], 17, -1473231341);
    b = md5_ff(b, c, d, a, x[i+ 7], 22, -45705983);
    a = md5_ff(a, b, c, d, x[i+ 8], 7 ,  1770035416);
    d = md5_ff(d, a, b, c, x[i+ 9], 12, -1958414417);
    c = md5_ff(c, d, a, b, x[i+10], 17, -42063);
    b = md5_ff(b, c, d, a, x[i+11], 22, -1990404162);
    a = md5_ff(a, b, c, d, x[i+12], 7 ,  1804603682);
    d = md5_ff(d, a, b, c, x[i+13], 12, -40341101);
    c = md5_ff(c, d, a, b, x[i+14], 17, -1502002290);
    b = md5_ff(b, c, d, a, x[i+15], 22,  1236535329);
    a = md5_gg(a, b, c, d, x[i+ 1], 5 , -165796510);
    d = md5_gg(d, a, b, c, x[i+ 6], 9 , -1069501632);
    c = md5_gg(c, d, a, b, x[i+11], 14,  643717713);
    b = md5_gg(b, c, d, a, x[i+ 0], 20, -373897302);
    a = md5_gg(a, b, c, d, x[i+ 5], 5 , -701558691);
    d = md5_gg(d, a, b, c, x[i+10], 9 ,  38016083);
    c = md5_gg(c, d, a, b, x[i+15], 14, -660478335);
    b = md5_gg(b, c, d, a, x[i+ 4], 20, -405537848);
    a = md5_gg(a, b, c, d, x[i+ 9], 5 ,  568446438);
    d = md5_gg(d, a, b, c, x[i+14], 9 , -1019803690);
    c = md5_gg(c, d, a, b, x[i+ 3], 14, -187363961);
    b = md5_gg(b, c, d, a, x[i+ 8], 20,  1163531501);
    a = md5_gg(a, b, c, d, x[i+13], 5 , -1444681467);
    d = md5_gg(d, a, b, c, x[i+ 2], 9 , -51403784);
    c = md5_gg(c, d, a, b, x[i+ 7], 14,  1735328473);
    b = md5_gg(b, c, d, a, x[i+12], 20, -1926607734);
    a = md5_hh(a, b, c, d, x[i+ 5], 4 , -378558);
    d = md5_hh(d, a, b, c, x[i+ 8], 11, -2022574463);
    c = md5_hh(c, d, a, b, x[i+11], 16,  1839030562);
    b = md5_hh(b, c, d, a, x[i+14], 23, -35309556);
    a = md5_hh(a, b, c, d, x[i+ 1], 4 , -1530992060);
    d = md5_hh(d, a, b, c, x[i+ 4], 11,  1272893353);
    c = md5_hh(c, d, a, b, x[i+ 7], 16, -155497632);
    b = md5_hh(b, c, d, a, x[i+10], 23, -1094730640);
    a = md5_hh(a, b, c, d, x[i+13], 4 ,  681279174);
    d = md5_hh(d, a, b, c, x[i+ 0], 11, -358537222);
    c = md5_hh(c, d, a, b, x[i+ 3], 16, -722521979);
    b = md5_hh(b, c, d, a, x[i+ 6], 23,  76029189);
    a = md5_hh(a, b, c, d, x[i+ 9], 4 , -640364487);
    d = md5_hh(d, a, b, c, x[i+12], 11, -421815835);
    c = md5_hh(c, d, a, b, x[i+15], 16,  530742520);
    b = md5_hh(b, c, d, a, x[i+ 2], 23, -995338651);
    a = md5_ii(a, b, c, d, x[i+ 0], 6 , -198630844);
    d = md5_ii(d, a, b, c, x[i+ 7], 10,  1126891415);
    c = md5_ii(c, d, a, b, x[i+14], 15, -1416354905);
    b = md5_ii(b, c, d, a, x[i+ 5], 21, -57434055);
    a = md5_ii(a, b, c, d, x[i+12], 6 ,  1700485571);
    d = md5_ii(d, a, b, c, x[i+ 3], 10, -1894986606);
    c = md5_ii(c, d, a, b, x[i+10], 15, -1051523);
    b = md5_ii(b, c, d, a, x[i+ 1], 21, -2054922799);
    a = md5_ii(a, b, c, d, x[i+ 8], 6 ,  1873313359);
    d = md5_ii(d, a, b, c, x[i+15], 10, -30611744);
    c = md5_ii(c, d, a, b, x[i+ 6], 15, -1560198380);
    b = md5_ii(b, c, d, a, x[i+13], 21,  1309151649);
    a = md5_ii(a, b, c, d, x[i+ 4], 6 , -145523070);
    d = md5_ii(d, a, b, c, x[i+11], 10, -1120210379);
    c = md5_ii(c, d, a, b, x[i+ 2], 15,  718787259);
    b = md5_ii(b, c, d, a, x[i+ 9], 21, -343485551);
    a = safe_add(a, olda);
    b = safe_add(b, oldb);
    c = safe_add(c, oldc);
    d = safe_add(d, oldd);
  }
  return Array(a, b, c, d);
}
function md5_ff(a, b, c, d, x, s, t){
  return md5_cmn((b & c) | ((~b) & d), a, b, x, s, t);}
function md5_gg(a, b, c, d, x, s, t){
  return md5_cmn((b & d) | (c & (~d)), a, b, x, s, t);}
function md5_hh(a, b, c, d, x, s, t){
  return md5_cmn(b ^ c ^ d, a, b, x, s, t);}
function md5_ii(a, b, c, d, x, s, t){
  return md5_cmn(c ^ (b | (~d)), a, b, x, s, t);}
function md5_cmn(q, a, b, x, s, t){
  return safe_add(bit_rol(safe_add(safe_add(a, q), safe_add(x, t)), s),b);}
function bit_rol(num, cnt){
  return (num << cnt) | (num >>> (32 - cnt));}
function safe_add(x, y){
  var lsw = (x & 0xFFFF) + (y & 0xFFFF);
  var msw = (x >> 16) + (y >> 16) + (lsw >> 16);
  return (msw << 16) | (lsw & 0xFFFF);
}
function str2binl(str)
{
  var bin = Array();
  var mask = (1 << chrsz) - 1;
  for(var i = 0; i < str.length * chrsz; i += chrsz)
    bin[i>>5] |= (str.charCodeAt(i / chrsz) & mask) << (i%32);
  return bin;
}
function binl2hex(binarray)
{
  var hex_tab = hexcase ? '0123456789ABCDEF' : '0123456789abcdef';
  var str = '';
  for(var i = 0; i < binarray.length * 4; i++)
    str += hex_tab.charAt((binarray[i>>2] >> ((i%4)*8+4)) & 0xF) +
           hex_tab.charAt((binarray[i>>2] >> ((i%4)*8  )) & 0xF);
  return str;
}
// PHPAuthentication login function
function login(){
  chg = '". $this->challenge() ."';
  pwd = document.forms[0].password.value;
  document.forms[0].password.value = '';
  document.forms[0].". $this->parname .".value=md5(chg+pwd);
}
";
  }
};

?>
